# Generated by Django 4.2.5 on 2023-09-07 19:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('oauth2_provider', '0008_alter_accesstoken_token'),
    ]

    operations = [
        migrations.AddField(
            model_name='application',
            name='hash_client_secret',
            field=models.BooleanField(default=True),
        ),
    ]
