# Generated by Django 4.2.4 on 2023-09-11 07:03

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("oauth2_provider", "0007_application_post_logout_redirect_uris"),
    ]

    operations = [
        migrations.AlterField(
            model_name="accesstoken",
            name="token",
            field=models.CharField(db_index=True, max_length=255, unique=True),
        ),
    ]
