# Generated by Django 4.1.5 on 2023-09-27 20:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("oauth2_provider", "0009_add_hash_client_secret"),
    ]

    operations = [
        migrations.AddField(
            model_name="application",
            name="allowed_origins",
            field=models.TextField(
                blank=True,
                help_text="Allowed origins list to enable CORS, space separated",
                default="",
            ),
        ),
    ]
